unit Print_F;
{
  :   㧨 ..    2:5080/37.5@fidonet.org
  : 㭮 ..  2:5080/37.14@fidonet.org
}
interface

function PrintF(Format: string; Params: array of Pointer): string;
(*
  PrintF  墠  ᪠  C-㭪樨.
     ଠ஢ ப, 祬 FormatStr.

  䨪樨 ࠬ஢:

  %[-][0][nnn][.mmm]{ c | d | [l]i | [l]f | s | [l]x }

  
  '-' - ࠢ ;
  '0' -  ﬨ;
  nnn - ਭ ;
  mmm - 樨 ᫥ 窨;
  'c' -  ᨬ;
  'd' - ;
  'i' - 楫 ᫮ (word, integer);
  'f' - ⢥ ᫮ (real);
  's' - ப;
  'x' - ⭠筮 ।⠢ (word, integer).
  䨪 'l' 砥  筮 (longint, double).
*)
function Space(Len: byte): string;
function Replicate(const R: string; Len: byte): string;

implementation

uses TDate, Dos;

type
  Long = record
    LowWord, HighWord: Word;
  end;

const
  Digits: array[0..$F] of Char = '0123456789ABCDEF';

function HexW(W: Word): string;
begin
  HexW[0] := #4;
  HexW[1] := Digits[hi(W) shr 4];
  HexW[2] := Digits[hi(W) and $F];
  HexW[3] := Digits[lo(W) shr 4];
  HexW[4] := Digits[lo(W) and $F];
end;

function HexL(L: LongInt): string;
begin
  with Long(L) do
    HexL := HexW(HighWord)+HexW(LowWord);
end;

function PrintF(Format: string; Params: array of Pointer): string;

var
  Ind: integer;
  S, Wstr: string;
  LongPending,
  ParmPending,
  PointPending,
  LeftJustify,
  ZeroPad: boolean;
  ParmIndex: integer;
  Width, Decimals: integer;
  CurrChar: char;

  procedure ClearFlags;
  begin
    LongPending := False;
    ParmPending := False;
    PointPending := False;
    LeftJustify := False;
    ZeroPad := False;
    Width := 0;
    Decimals := 0;
  end;

  function NextChar: char;
  begin
    NextChar := ' ';
    if Ind <= Length(Format) then
    begin
      Inc(Ind);
      NextChar := Format[Ind];
    end;
  end;

  function PrevChar: char;
  begin
    if Ind > 1 then PrevChar := Format[Ind - 1]
      else PrevChar := Format[Ind];
  end;

  procedure SaveChar(C: char);
  begin
    S := S + C;
  end;

  procedure SaveString(Wstr: string);
  var
    i: integer;
    PadChar: char;
    Delta: integer;
  begin
    if ZeroPad then PadChar := '0' else PadChar := ' ';
    if Length(Wstr) > Width then Width := Length(Wstr);
    Delta := Width - Length(Wstr);

    for i := 1 to Width do
      if LeftJustify then
        if i <= Length(Wstr)
          then SaveChar(Wstr[i])
          else SaveChar(PadChar)
      else if i <= Delta
        then SaveChar(PadChar)
        else if Delta > 0
          then SaveChar(Wstr[i-Delta])
          else SaveChar(Wstr[i]);
  end;

  procedure Convert(Base: char);
  begin
    Inc(ParmIndex);
    if Params[ParmIndex] = nil then
    { substitute null parameter with spaces }
    case UpCase(Base) of
      'C': Wstr := ' ';
      'D': if Width < 10 then Wstr := Space(8) else Wstr := Space(10);
      'X': if LongPending then Wstr := Space(8) else Wstr := Space(4);
      else Wstr := Space(Width);
    end else
    case UpCase(Base) of
      'C': begin
             Wstr[0] := #1;
             Wstr[1] := Char(Params[ParmIndex]^);
           end;
      'D': if Width < 10
             then Wstr := DateToDateString(Date(Params[ParmIndex]^), False)
             else Wstr := DateToDateString(Date(Params[ParmIndex]^), True);
      'I': if LongPending
             then Str(Longint(Params[ParmIndex]^):-Width, Wstr)
             else Str(Integer(Params[ParmIndex]^):-Width, Wstr);
      'F': begin
             if LongPending
               then Str(Double(Params[ParmIndex]^):-Width:Decimals, Wstr)
               else Str(Real(Params[ParmIndex]^):-Width:Decimals, Wstr);
           end;
      'X': if LongPending
             then Wstr := HexL(Longint(Params[ParmIndex]^))
             else Wstr := HexW(Integer(Params[ParmIndex]^));
      'S': if Decimals > 0
             then Wstr := Copy(String(Params[ParmIndex]^), 1, Decimals)
             else if Width = 0 then Wstr := String(Params[ParmIndex]^)
               else Wstr := Copy(String(Params[ParmIndex]^), 1, Width);
    end;
    SaveString(Wstr);
    ClearFlags;
  end;

  procedure CountNumbers(C: char);
  begin
    if PointPending
      then Decimals := Decimals * 10 + Ord(C) - 48
      else Width := Width * 10 + Ord(C) - 48;
  end;

begin
  S := '';
  Ind := 0;
  ParmIndex := -1;
  ClearFlags;
  while Ind < Length(Format) do
  begin
    CurrChar := NextChar;
    if CurrChar = '%' then ParmPending := True;
    if ParmPending
    then begin
      if CurrChar = '%' then
        if Ind < Length(Format) then CurrChar := NextChar
          else CurrChar := ' ';
      case UpCase(CurrChar) of
        '%': begin
               ClearFlags;
               SaveChar(CurrChar);
             end;
        'L': LongPending := True;
        'C', 'D', 'I', 'F', 'X', 'S': Convert(CurrChar);
        '0'..'9':
          begin
            if CurrChar = '0' then
              if not (PrevChar in ['%', '-']) then CountNumbers(CurrChar)
                else ZeroPad := True
              else CountNumbers(CurrChar);
          end;
        '-': LeftJustify := True;
        '.': PointPending := True;
        else begin
          ClearFlags;
          SaveChar(CurrChar);
        end;
      end;
    end
    else SaveChar(CurrChar);
  end;

  PrintF := S;
end;

function Space(Len: byte): string;
begin
  Space := Replicate(' ', Len);
end;

function Replicate(const R: string; Len: byte): string;
var
  S: string;
  i: byte;
begin
  S := '';
  for i := 1 to Len do S := S + R;
  Replicate := S;
end;

end.
