unit TDate;
{
  Implements Date type.
}

interface

type
  Date = LongInt;
  DateString = string[10];

const
  CDelimiter   = '.';

  MinYear = 0000;
  MaxYear = 9999;
  MinDate = $00000000;     {= 01.01.1600}
  MaxDate = $002EC723;     {= 31.12.3999}
  BadDate = $FFFFFFFF;
  First2Months = 59;       {1600 was a leap year}

  Century: word = 1900;

function DaysInMonth(Month, Year: Integer): Integer;
function DateStringIsBlank(S: string): boolean;
function ValidDate(Day, Month, Year: Integer): Boolean;
function DMYtoDate(Day, Month, Year: Integer): Date;
procedure DateToDMY(Julian: Date; var Day, Month, Year: Integer);
function DMYtoDateString(D, M, Y: Integer; FullYear: boolean): DateString;
function DateStringToDMY(S: DateString; var D, M, Y: Integer): Boolean;
function DateStringToDate(S: DateString): Date;
function DateToDateString(Julian: Date; FullYear: boolean): DateString;
function Today: Date;
function IncDate(Julian: Date; Days, Months, Years: Integer): Date;
function DMYtoString(D, M, Y: Integer): string;
function DateToString(Julian: Date): string;

implementation

uses Dos;

var
  M, D, Y: Integer;

function IsLeapYear(Year: Integer): Boolean;
begin
  IsLeapYear := (Year mod 4 = 0) and (Year mod 4000 <> 0) and
    ((Year mod 100 <> 0) or (Year mod 400 = 0));
end;

function DaysInMonth(Month, Year: Integer): Integer;
begin
  if Word(Year) < 100 then Inc(Year, Century);
  case Month of
    1, 3, 5, 7, 8, 10, 12: DaysInMonth := 31;
    4, 6, 9, 11: DaysInMonth := 30;
    2: DaysInMonth := 28+Ord(IsLeapYear(Year));
  else DaysInMonth := 0;
  end;
end;

function DateStringIsBlank(S: string): boolean;
var
  i, j: integer;
begin
  j := 0;
  for i := 1 to Length(S) do
    if not (S[i] in [' ', CDelimiter]) then Inc(j);
  DateStringIsBlank := j = 0;
end;

function ValidDate(Day, Month, Year : Integer) : Boolean;
begin
  if (Day < 1) or (Year < MinYear) or (Year > MaxYear) then
    ValidDate := False
  else case Month of
    1..12: ValidDate := Day <= DaysInMonth(Month, Year);
  else ValidDate := False;
  end
end;

function DMYtoDate(Day, Month, Year: Integer): Date;
begin
  if not ValidDate(Day, Month, Year) then DMYtoDate := BadDate
  else if (Year = MinYear) and (Month < 3) then
    if Month = 1 then
      DMYtoDate := Pred(Day)
    else
      DMYtoDate := Day+30
  else begin
    if Month > 2 then
      Dec(Month, 3)
    else begin
      Inc(Month, 9);
      Dec(Year);
    end;
    Dec(Year, MinYear);
    DMYtoDate :=
        ((LongInt(Year div 100)*146097) div 4)+
        ((LongInt(Year mod 100)*1461) div 4)+
        (((153*Month)+2) div 5)+Day+First2Months;
  end;
end;

procedure DateToDMY(Julian: Date; var Day, Month, Year: Integer);
var
  I, J : LongInt;
begin
  if Julian = BadDate then begin
    Day := 0;
    Month := 0;
    Year := 0;
  end
  else if Julian <= First2Months then begin
    Year := MinYear;
    if Julian <= 30 then begin
      Month := 1;
      Day := Succ(Julian);
    end
    else begin
      Month := 2;
      Day := Julian-30;
    end;
  end
  else begin
    I := (4*LongInt(Julian-First2Months))-1;
    J := (4*((I mod 146097) div 4))+3;
    Year := (100*(I div 146097))+(J div 1461);
    I := (5*(((J mod 1461)+4) div 4))-3;
    Month := I div 153;
    Day := ((I mod 153)+5) div 5;
    if Month < 10 then
      Inc(Month, 3)
    else begin
      Dec(Month, 9);
      Inc(Year);
    end;
    Inc(Year, MinYear);
  end;
end;

function DMYtoDateString(D, M, Y: Integer; FullYear: boolean): DateString;
var
  i: integer;
  S, W: string[10];
begin
  if Word(Y) < 100 then Inc(Y, Century);

  S := '';
  for i := 1 to 3 do
  begin
    case i of
      2: D := M;
      3: D := Y;
    end;
    Str(D:4, W);
    if i < 3 then Delete(W, 1, 2)
      else if not FullYear then Delete(W, 1, 2);
    if W[1] = ' ' then W[1] := '0';
    S := S + W + CDelimiter;
  end;
  if S[Length(S)] = CDelimiter then Delete(S, Length(S), 1);

  DMYtoDateString := S;
end;

function DateStringToDMY(S: DateString; var D, M, Y: Integer): Boolean;
var
  W: string[4];
  V, i, j: integer;
begin
  for j := 1 to 3 do
  begin
    i := Pos(CDelimiter, S);
    if i = 0 then i := 11;
    W := Copy(S, 1, i - 1);
    Delete(S, 1, i);
    Val(W, V, i);
    if i <> 0 then V := -1;
    case j of
      1: D := V;
      2: M := V;
      3:
        begin
          if length(W) <= 2 then Inc(V, Century);
          Y := V
        end;
    end;
  end;
  DateStringToDMY := ValidDate(D, M, Y);
end;

function DateStringToDate(S: DateString): Date;
begin
  if DateStringToDMY(S, D, M, Y) then DateStringToDate := DMYtoDate(D, M, Y)
    else DateStringToDate := BadDate;
end;

function DateToDateString(Julian: Date; FullYear: boolean): DateString;
begin
  if Julian = BadDate then DateToDateString := ''
  else begin
    DateToDMY(Julian, D, M, Y);
    DateToDateString := DMYtoDateString(D, M, Y, FullYear);
  end;
end;

function Today: Date;
var
  Year, Month, Day, DoW: Word;
begin
  GetDate(Year, Month, Day, DoW);
  Today := DMYtoDate(Day, Month, Year);
end;

function IncDate(Julian: Date; Days, Months, Years: Integer): Date;
var
  Day, Month, Year, Day28Delta: Integer;
begin
  DateToDMY(Julian, Day, Month, Year);
  Day28Delta := Day - 28;
  if Day28Delta < 0 then
    Day28Delta := 0
  else
    Day := 28;

  Inc(Year, Years);
  Inc(Year, Months div 12);
  Inc(Month, Months mod 12);
  if Month < 1 then begin
    Inc(Month, 12);
    Dec(Year);
  end
  else if Month > 12 then begin
    Dec(Month, 12);
    Inc(Year);
  end;

  Julian := DMYtoDate(Day, Month, Year);
  if Julian <> BadDate then begin
    Inc(Julian, Days);
    Inc(Julian, Day28Delta);
  end;
  IncDate := Julian;
end;

function DMYtoString(D, M, Y: Integer): string;
const
  Months: array[1..12] of string[8] =
  ('ﭢ', '䥢ࠫ', '', '५', '', '',
   '', '', 'ᥭ', '', '', '');
var
  s1: string[2];
  s2: string[4];
begin
  Str(D, s1);
  Str(Y, s2);
  DMYtoString := s1 + ' ' + Months[M] + ' ' + s2;
end;

function DateToString(Julian: Date): string;
var
  D, M, Y: integer;
begin
  DateToDMY(Julian, D, M, Y);
  DateToString := DMYtoString(D, M, Y);
end;

end.